import json
from inspect import ismethod
from typing import Optional, Union

from discord import Colour, Embed

from ..exceptions import BadColourArgument, EmbedParseError
from ..interface import Block
from ..interpreter import Context
from .helpers import helper_split, implicit_bool


def string_to_color(argument: str) -> Colour:
    arg = argument.replace("0x", "").lower()

    if arg[0] == "#":
        arg = arg[1:]
    try:
        value = int(arg, base=16)
        if not (0 <= value <= 0xFFFFFF):
            raise BadColourArgument(arg)
        return Colour(value=value)
    except ValueError:
        arg = arg.replace(" ", "_")
        method = getattr(Colour, arg, None)
        if arg.startswith("from_") or method is None or not ismethod(method):
            raise BadColourArgument(arg)
        return method()


def set_color(embed: Embed, attribute: str, value: str):
    value = string_to_color(value)
    setattr(embed, attribute, value)


def set_dynamic_url(embed: Embed, attribute: str, value: str):
    method = getattr(embed, f"set_{attribute}")
    method(url=value)


def add_field(embed: Embed, _: str, payload: str):
    try:
        name, value, _inline = helper_split(payload, 3)
        inline = implicit_bool(_inline)
        if inline is None:
            raise EmbedParseError(
                f"`inline` argument for `add_field` is not a boolean value (_inline)"
            )
    except ValueError:
        try:
            name, value = helper_split(payload, 2)
        except ValueError as exc:
            raise EmbedParseError("`add_field` payload was not split by |") from exc
        inline = False
    embed.add_field(name=name, value=value, inline=inline)


class EmbedBlock(Block):
    """
    An embed block will send an embed in the tag response.
    There are two ways to use the embed block, either by using properly
    formatted embed JSON from an embed generator or manually inputting
    the accepted embed attributes.

    **JSON**

    Using JSON to create an embed offers complete embed customization.
    Multiple embed generators are available online to visualize and generate
    embed JSON.

    **Usage:** ``{embed(<json>)}``

    **Payload:** None

    **Parameter:** json

    .. tagscript:: 
        {embed({"title":"Hello!", "description":"This is a test embed."})}
        {embed({
            "title":"Here's a random duck!",
            "image":{"url":"https://random-d.uk/api/randomimg"},
            "color":15194415
        })}

    **Manual**

    The following embed attributes can be set manually:

    *   ``title``
    *   ``description``
    *   ``color``
    *   ``url``
    *   ``thumbnail``
    *   ``image``
    *   ``field`` - (See below)

    Adding a field to an embed requires the payload to be split by ``|``, into
    either 2 or 3 parts. The first part is the name of the field, the second is
    the text of the field, and the third optionally specifies whether the field
    should be inline.

    **Usage:** ``{embed(<attribute>):<value>}``

    **Payload:** value

    **Parameter:** attribute

    .. tagscript:: 
        {embed(color):#37b2cb}
        {embed(title):Rules}
        {embed(description):Follow these rules to ensure a good experience in our server!}
        {embed(field):Rule 1|Respect everyone you speak to.|false}

    Both methods can be combined to create an embed in a tag.
    The following tagscript uses JSON to create an embed with fields and later
    set the embed title.

    ::

        {embed({{"fields":[{"name":"Field 1","value":"field description","inline":false}]})}
        {embed(title):my embed title}
    """

    ACCEPTED_NAMES = ("embed",)

    ATTRIBUTE_HANDLERS = {
        "description": setattr,
        "title": setattr,
        "color": set_color,
        "colour": set_color,
        "url": setattr,
        "thumbnail": set_dynamic_url,
        "image": set_dynamic_url,
        "field": add_field,
    }

    @staticmethod
    def get_embed(ctx: Context) -> Embed:
        return ctx.response.actions.get("embed", Embed())

    @staticmethod
    def value_to_color(value: Optional[Union[int, str]]) -> Colour:
        if value is None or isinstance(value, Colour):
            return value
        if isinstance(value, int):
            return Colour(value)
        elif isinstance(value, str):
            return string_to_color(value)
        else:
            raise EmbedParseError("Received invalid type for color key (expected string or int)")

    def text_to_embed(self, text: str) -> Embed:
        try:
            data = json.loads(text)
        except json.decoder.JSONDecodeError as error:
            raise EmbedParseError(error) from error

        if data.get("embed"):
            data = data["embed"]
        if data.get("timestamp"):
            data["timestamp"] = data["timestamp"].strip("Z")

        color = data.pop("color", data.pop("colour", None))

        try:
            embed = Embed.from_dict(data)
        except Exception as error:
            raise EmbedParseError(error) from error
        else:
            if color := self.value_to_color(color):
                embed.color = color
            return embed

    @classmethod
    def update_embed(cls, embed: Embed, attribute: str, value: str) -> Embed:
        handler = cls.ATTRIBUTE_HANDLERS[attribute]
        try:
            handler(embed, attribute, value)
        except Exception as error:
            raise EmbedParseError(error) from error
        return embed

    @staticmethod
    def return_error(error: Exception) -> str:
        return f"Embed Parse Error: {error}"

    @staticmethod
    def return_embed(ctx: Context, embed: Embed) -> str:
        try:
            length = len(embed)
        except Exception as error:
            return str(error)
        if length > 6000:
            return f"`MAX EMBED LENGTH REACHED ({length}/6000)`"
        ctx.response.actions["embed"] = embed
        return ""

    def process(self, ctx: Context) -> Optional[str]:
        if not ctx.verb.parameter:
            return self.return_embed(ctx, self.get_embed(ctx))

        lowered = ctx.verb.parameter.lower()
        try:
            if ctx.verb.parameter.startswith("{") and ctx.verb.parameter.endswith("}"):
                embed = self.text_to_embed(ctx.verb.parameter)
            elif lowered in self.ATTRIBUTE_HANDLERS and ctx.verb.payload:
                embed = self.get_embed(ctx)
                embed = self.update_embed(embed, lowered, ctx.verb.payload)
            else:
                return
        except EmbedParseError as error:
            return self.return_error(error)

        return self.return_embed(ctx, embed)



from typing import Optional

from ..interface import verb_required_block
from ..interpreter import Context


class SubstringBlock(verb_required_block(True, parameter=True)):
    ACCEPTED_NAMES = ("substr", "substring")

    def process(self, ctx: Context) -> Optional[str]:
        try:
            if "-" not in ctx.verb.parameter:
                return ctx.verb.payload[int(float(ctx.verb.parameter)) :]

            spl = ctx.verb.parameter.split("-")
            start = int(float(spl[0]))
            end = int(float(spl[1]))
            return ctx.verb.payload[start:end]
        except:
            return



from typing import Optional

from ..exceptions import StopError
from ..interface import verb_required_block
from ..interpreter import Context
from . import helper_parse_if


class StopBlock(verb_required_block(True, parameter=True)):
    """
    The stop block stops tag processing if the given parameter is true.
    If a message is passed to the payload it will return that message.

    **Usage:** ``{stop(<bool>):[string]}``

    **Aliases:** ``halt, error``

    **Payload:** string, None

    **Parameter:** bool

    **Example:**

    .. tagscript::

        {stop(=={args}):You must provide arguments for this tag.}
        enforces providing arguments for a tag
    """

    ACCEPTED_NAMES = ("stop", "halt", "break")

    def process(self, ctx: Context) -> Optional[str]:
        if helper_parse_if(ctx.verb.parameter):
            raise StopError("" if ctx.verb.payload is None else ctx.verb.payload)
        return ""



from discord import Guild
from re import match

from ..adapter import discordadapters
from ..interface import verb_required_block
from ..interpreter import Context\


"""
I have to rethink this later"""


class GetDiscord(verb_required_block(True, payload=True, parameter=True)):
    """
    This block allows users to get or fetch discord objects such as users, channels, etc.

    This will set the target block with new attributes


    **Usage:** ``{request(["Any Target attribute"]):<string>}``

    **Payload:** request_type|query

    Request type can be any of the following: "channel", "user", query is just what you're searching for

    **Parameter:** "channel", "user"

    **Example:**
    
    .. tagscript::

        {request(user):user|360061101477724170}
        _Leg3ndary#5759
    """

    ACCEPTED_NAMES = ("request", "getdiscord")

    def __init__(self, guild: Guild, limit: int = 3, allow_fetch: bool = True) -> None:
        """
        You need access to a guild object from dpy for this to work.


        ``limit`` is how many blocks are allowed to be used in a single tag
        
        ``allow_fetch`` tells us whether we're allowed to fetch users, or use cache only.

        If set to ``False``, and you're cache is disabled, this block is essentially useless.
        """
        super().__init__()
        self.guild = guild
        self.limit = limit
        self.allow_fetch = allow_fetch

    async def process(self, ctx: Context):
        if "|" not in ctx.verb.payload:
            return f"`YOU DON'T SEEM TO HAVE A PIPE (|)`"
        _type = ctx.verb.payload.split("|")[0].lower()
        query = ctx.verb.payload.split("|")[1]

        if query.isdigit and match("[0-9]{15,19}", query):

            actions = ctx.response.actions.get("request")
            
            if actions:
                if len(actions) > self.limit:
                    return f"`REQUEST LIMIT REACHED ({self.limit})`"
            
            else:
                actions = []
            
            if _type == "channel":
                channel = self.guild.get_channel(int(query)) or (await self.guild.fetch_channel(int(query)))
                if channel:
                    channel = discordadapters.ChannelAdapter(channel)
                else:
                    return f"Channel `{query}` not found"
                    
            elif _type =="user":
                user = self.guild.get_member(int(query)) or (await self.guild.fetch_member(int(query)))
                if user:
                    user = discordadapters.MemberAdapter()
                else:
                    return f"User `{query}` not found"
            else:
                return f"Type {_type} not found"

            ctx.response.actions["request"] = actions.append(f"{_type}:{query}")



from typing import Optional

from ..interface import Block
from ..interpreter import Context
from ..verb import Verb


class ShortCutRedirectBlock(Block):
    def __init__(self, var_name):
        self.redirect_name = var_name

    def will_accept(self, ctx: Context) -> bool:
        return ctx.verb.declaration.isdigit()

    def process(self, ctx: Context) -> Optional[str]:
        blank = Verb()
        blank.declaration = self.redirect_name
        blank.parameter = ctx.verb.declaration
        ctx.verb = blank
        return None



from typing import Optional

from ..interface import Block
from ..interpreter import Context


class LooseVariableGetterBlock(Block):
    """
    The loose variable block represents the adapters for any seeded or defined variables.
    This variable implementation is considered "loose" since it checks whether the variable is
    valid during :meth:`process`, rather than :meth:`will_accept`.

    **Usage:** ``{<variable_name>([parameter]):[payload]}``

    **Aliases:** This block is valid for any inputted declaration.

    **Payload:** Depends on the variable's underlying adapter.

    **Parameter:** Depends on the variable's underlying adapter.

    **Examples:**
    
    .. tagscript::

        {=(example):This is my variable.}
        {example}
        This is my variable.
    """

    def will_accept(self, ctx: Context) -> bool:
        return True

    def process(self, ctx: Context) -> Optional[str]:
        if ctx.verb.declaration in ctx.response.variables:
            return ctx.response.variables[ctx.verb.declaration].get_value(ctx.verb)
        else:
            return None



from __future__ import annotations

from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from ..interpreter import Context


class Adapter:
    """
    The base class for TagScript blocks.

    Implementations must subclass this to create adapters.
    """

    def __repr__(self):
        return f"<{type(self).__qualname__} at {hex(id(self))}>"

    def get_value(self, ctx: Context) -> Optional[str]:
        """
        Processes the adapter's actions for a given :class:`~bTagScript.interpreter.Context`.

        Subclasses must implement this.

        Parameters
        ----------
        ctx: Context
            The context object containing the TagScript :class:`~bTagScript.verb.Verb`.

        Returns
        -------
        Optional[str]
            The adapters's processed value.

        Raises
        ------
        NotImplementedError
            The subclass did not implement this required method.
        """
        raise NotImplementedError



from typing import Optional

__all__ = ("Verb",)


class Verb:
    """
    Represents the passed TagScript block.

    Parameters
    ----------
    verb_string: Optional[str]
        The string to parse into a verb.
    limit: int
        The maximum number of characters to parse.
    dot_parameter: bool
        Whether the parameter should be followed after a "." or use the default of parantheses.

    Attributes
    ----------
    declaration: Optional[str]
        The text used to declare the block.
    parameter: Optional[str]
        The text passed to the block parameter in the parentheses.
    payload: Optional[str]
        The text passed to the block payload after the colon.

    Example
    -------
    Below is a visual representation of a block and its attributes::

        {declaration(parameter):payload}

        # dot_parameter = True
        {declaration.parameter:payload}
    """

    __slots__ = (
        "declaration",
        "parameter",
        "payload",
        "parsed_string",
        "dec_depth",
        "dec_start",
        "skip_next",
        "parsed_length",
        "dot_parameter",
    )

    def __init__(
        self, verb_string: Optional[str] = None, *, limit: int = 2000, dot_parameter: bool = False
    ):
        self.declaration: Optional[str] = None
        self.parameter: Optional[str] = None
        self.payload: Optional[str] = None
        self.dot_parameter = dot_parameter
        if verb_string is None:
            return
        self.__parse(verb_string, limit)

    def __str__(self):
        """This makes Verb compatible with str(x)"""
        response = "{"
        if self.declaration is not None:
            response += self.declaration
        if self.parameter is not None:
            response += f".{self.parameter}" if self.dot_parameter else f"({self.parameter})"
        if self.payload is not None:
            response += ":" + self.payload
        return response + "}"

    def __repr__(self):
        attrs = ("declaration", "parameter", "payload")
        inner = " ".join(f"{attr}={getattr(self, attr)!r}" for attr in attrs)
        return f"<Verb {inner}>"

    def __parse(self, verb_string: str, limit: int):
        self.parsed_string = verb_string[1:-1][:limit]
        self.parsed_length = len(self.parsed_string)
        self.dec_depth = 0
        self.dec_start = 0
        self.skip_next = False

        parse_parameter = (
            self._parse_dot_parameter if self.dot_parameter else self._parse_paranthesis_parameter
        )

        for i, v in enumerate(self.parsed_string):
            if self.skip_next:
                self.skip_next = False
                continue
            elif v == "\\":
                self.skip_next = True
                continue

            if v == ":" and not self.dec_depth:
                # if v == ":" and not dec_depth:
                self.set_payload()
                return
            elif parse_parameter(i, v):
                return
        else:
            self.set_payload()

    def _parse_paranthesis_parameter(self, i: int, v: str) -> bool:
        if v == "(":
            self.open_parameter(i)
        elif v == ")" and self.dec_depth:
            return self.close_parameter(i)
        return False

    def _parse_dot_parameter(self, i: int, v: str) -> bool:
        if v == ".":
            self.open_parameter(i)
        elif (v == ":" or i == self.parsed_length - 1) and self.dec_depth:
            return self.close_parameter(i + 1)
        return False

    def set_payload(self):
        res = self.parsed_string.split(":", 1)
        if len(res) == 2:
            self.payload = res[1]
        self.declaration = res[0]

    def open_parameter(self, i: int):
        self.dec_depth += 1
        if not self.dec_start:
            self.dec_start = i
            self.declaration = self.parsed_string[:i]

    def close_parameter(self, i: int) -> bool:
        self.dec_depth -= 1
        if self.dec_depth == 0:
            self.parameter = self.parsed_string[self.dec_start + 1 : i]
            try:
                if self.parsed_string[i + 1] == ":":
                    self.payload = self.parsed_string[i + 2 :]
            except IndexError:
                pass
            return True
        return False



from typing import Optional

from ..adapter import StringAdapter
from ..interface import verb_required_block
from ..interpreter import Context


class AssignmentBlock(verb_required_block(False, parameter=True)):
    """
    Variables are useful for choosing a value and referencing it later in a tag.
    Variables can be referenced using brackets as any other block.

    **Usage:** ``{=(<name>):<value>}``

    **Aliases:** ``assign, let, var, =``

    **Payload:** value

    **Parameter:** name

    **Examples:**
    
    .. tagscript::

        {=(prefix):!}
        The prefix here is `{prefix}`.
        The prefix here is `!`.

        {assign(day):Monday}
        {if({day}==Wednesday):It's Wednesday my dudes!|The day is {day}.}
        The day is Monday.
    """

    ACCEPTED_NAMES = ("=", "assign", "let", "var")

    def process(self, ctx: Context) -> Optional[str]:
        if ctx.verb.parameter is None:
            return None
        ctx.response.variables[ctx.verb.parameter] = StringAdapter(str(ctx.verb.payload))
        return ""



from __future__ import annotations

from typing import TYPE_CHECKING

from discord.ext.commands import Cooldown

if TYPE_CHECKING:
    from .interpreter import Interpreter, Response


__all__ = (
    "TagScriptError",
    "WorkloadExceededError",
    "ProcessError",
    "EmbedParseError",
    "BadColourArgument",
    "StopError",
    "CooldownExceeded",
)


class TagScriptError(Exception):
    """Base class for all module errors."""


class WorkloadExceededError(TagScriptError):
    """Raised when the interpreter goes over its passed character limit."""


class ProcessError(TagScriptError):
    """
    Raised when an exception occurs during interpreter processing.

    Attributes
    ----------
    original: Exception
        The original exception that occurred during processing.
    response: Response
        The incomplete response that was being processed when the exception occurred.
    interpreter: Interpreter
        The interpreter used for processing.
    """

    def __init__(self, error: Exception, response: Response, interpreter: Interpreter):
        self.original: Exception = error
        self.response: Response = response
        self.interpreter: Interpreter = interpreter
        super().__init__(error)


class EmbedParseError(TagScriptError):
    """Raised if an exception occurs while attempting to parse an embed."""


class BadColourArgument(EmbedParseError):
    """
    Raised when the passed input fails to convert to `discord.Colour`.

    Attributes
    ----------
    argument: str
        The invalid input.
    """

    def __init__(self, argument: str):
        self.argument = argument
        super().__init__(f'Colour "{argument}" is invalid.')


class StopError(TagScriptError):
    """
    Raised by the StopBlock to stop processing.

    Attributes
    ----------
    message: str
        The stop error message.
    """

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)


class CooldownExceeded(StopError):
    """
    Raised by the cooldown block when a cooldown is exceeded.

    Attributes
    ----------
    message: str
        The cooldown error message.
    cooldown: discord.ext.commands.Cooldown
        The cooldown bucket with information on the cooldown.
    key: str
        The cooldown key that reached its cooldown.
    retry_after: float
        The seconds left til the cooldown ends.
    """

    def __init__(self, message: str, cooldown: Cooldown, key: str, retry_after: float):
        self.cooldown = cooldown
        self.key = key
        self.retry_after = retry_after
        super().__init__(message)



from typing import Optional

from ..interface import Block
from ..interpreter import Context


class StrictVariableGetterBlock(Block):
    """
    The strict variable block represents the adapters for any seeded or defined variables.
    This variable implementation is considered "strict" since it checks whether the variable is
    valid during :meth:`will_accept` and is only processed if the declaration refers to a valid
    variable.

    **Usage:** ``{<variable_name>([parameter]):[payload]}``

    **Aliases:** This block is valid for any variable name in `Response.variables`.

    **Payload:** Depends on the variable's underlying adapter.

    **Parameter:** Depends on the variable's underlying adapter.

    **Examples:**
    
    .. tagscript::

        {=(example):This is my variable.}
        {example}
        This is my variable.
    """

    def will_accept(self, ctx: Context) -> bool:
        return ctx.verb.declaration in ctx.response.variables

    def process(self, ctx: Context) -> Optional[str]:
        return ctx.response.variables[ctx.verb.declaration].get_value(ctx.verb)



from ..interface import Adapter
from ..verb import Verb


class IntAdapter(Adapter):
    __slots__ = ("integer",)

    def __init__(self, integer: int):
        self.integer: int = int(integer)

    def __repr__(self):
        return f"<{type(self).__qualname__} integer={repr(self.integer)}>"

    def get_value(self, ctx: Verb) -> str:
        return str(self.integer)



from random import choice

from discord import TextChannel, Thread

from ..interface import Adapter
from ..utils import escape_content
from ..verb import Verb

__all__ = (
    "AttributeAdapter",
    "MemberAdapter",
    "ChannelAdapter",
    "GuildAdapter",
)


class AttributeAdapter(Adapter):
    __slots__ = ("object", "_attributes", "_methods")

    def __init__(self, base):
        self.object = base
        self._attributes = {
            "id": base.id,
            "created_at": base.created_at,
            "timestamp": int(base.created_at.timestamp()),
            "name": getattr(base, "name", str(base)),
        }
        self._methods = {}
        self.update_attributes()
        self.update_methods()

    def __repr__(self):
        return f"<{type(self).__qualname__} object={self.object!r}>"

    def update_attributes(self):
        pass

    def update_methods(self):
        pass

    def get_value(self, ctx: Verb) -> str:
        should_escape = False

        if ctx.parameter is None:
            return_value = str(self.object)
        else:
            try:
                value = self._attributes[ctx.parameter]
            except KeyError:
                if method := self._methods.get(ctx.parameter):
                    value = method()
                else:
                    return

            if isinstance(value, tuple):
                value, should_escape = value

            return_value = str(value) if value is not None else None

        return escape_content(return_value) if should_escape else return_value


class MemberAdapter(AttributeAdapter):
    """
    The ``{author}`` block with no parameters returns the tag invoker's full username
    and discriminator, but passing the attributes listed below to the block payload
    will return that attribute instead.

    **Aliases:** ``user``

    **Usage:** ``{author([attribute])``

    **Payload:** None

    **Parameter:** attribute, None

    Attributes
    ----------
    id
        The author's Discord ID.
    name
        The author's username.
    nick
        The author's nickname, if they have one, else their username.
    avatar
        A link to the author's avatar, which can be used in embeds.
    discriminator
        The author's discriminator.
    created_at
        The author's account creation date.
    timestamp
        The author's account creation date as a UTC timestamp.
    joined_at
        The date the author joined the server.
    mention
        A formatted text that pings the author.
    bot
        Whether or not the author is a bot.
    color
        The author's top role's color as a hex code.
    top_role
        The author's top role.
    roleids
        A list of the author's role IDs, split by spaces.
    """

    def update_attributes(self):
        additional_attributes = {
            "color": self.object.color,
            "colour": self.object.color,
            "nick": self.object.display_name,
            "avatar": (self.object.avatar.url, False),
            "discriminator": self.object.discriminator,
            "joined_at": getattr(self.object, "joined_at", self.object.created_at),
            "mention": self.object.mention,
            "bot": self.object.bot,
            "top_role": getattr(self.object, "top_role", None),
        }
        if roleids := getattr(self.object, "_roles", None):
            additional_attributes["roleids"] = " ".join(str(r) for r in roleids)
        self._attributes.update(additional_attributes)


class ChannelAdapter(AttributeAdapter):
    """
    The ``{channel}`` block with no parameters returns the channel's full name
    but passing the attributes listed below to the block payload
    will return that attribute instead.

    **Usage:** ``{channel([attribute])``

    **Payload:** None

    **Parameter:** attribute, None

    Attributes
    ----------
    id
        The channel's ID.
    name
        The channel's name.
    created_at
        The channel's creation date.
    timestamp
        The channel's creation date as a UTC timestamp.
    nsfw
        Whether the channel is nsfw.
    mention
        A formatted text that pings the channel.
    topic
        The channel's topic.
    """

    def update_attributes(self) -> None:
        """
        Update block attributes
        """
        if isinstance(self.object, TextChannel):
            additional_attributes = {
                "channel_type": "textchannel",
                "nsfw": self.object.nsfw,
                "mention": self.object.mention,
                "topic": self.object.topic or None,
            }
            self._attributes.update(additional_attributes)
        elif isinstance(self.object, Thread):
            pass


class GuildAdapter(AttributeAdapter):
    """
    The ``{server}`` block with no parameters returns the server's name
    but passing the attributes listed below to the block payload
    will return that attribute instead.

    **Aliases:** ``guild``

    **Usage:** ``{server([attribute])``

    **Payload:** None

    **Parameter:** attribute, None

    Attributes
    ----------
    id
        The server's ID.
    name
        The server's name.
    icon
        A link to the server's icon, which can be used in embeds.
    created_at
        The server's creation date.
    timestamp
        The server's creation date as a UTC timestamp.
    member_count
        The server's member count.
    bots
        The number of bots in the server.
    humans
        The number of humans in the server.
    description
        The server's description if one is set, or "No description".
    random
        A random member from the server.
    """

    def update_attributes(self) -> None:
        """
        Update block attributes
        """
        guild = self.object
        bots = 0
        humans = 0
        for m in guild.members:
            if m.bot:
                bots += 1
            else:
                humans += 1
        member_count = guild.member_count
        additional_attributes = {
            "icon": guild.icon.url if guild.icon else "",
            "member_count": member_count,
            "members": member_count,
            "bots": bots,
            "humans": humans,
            "description": guild.description or "No description",
        }
        self._attributes.update(additional_attributes)

    def update_methods(self) -> None:
        """
        Update methods for the block
        """
        additional_methods = {"random": self.random_member}
        self._methods.update(additional_methods)

    def random_member(self) -> None:
        """
        Return a random member
        """
        return choice(self.object.members)



import time
from typing import Any, Dict, Optional

from discord.ext.commands import CooldownMapping

from ..exceptions import CooldownExceeded
from ..interface import verb_required_block
from ..interpreter import Context
from .helpers import helper_split

__all__ = ("CooldownBlock",)


class CooldownBlock(verb_required_block(True, payload=True, parameter=True)):
    """
    The cooldown block implements cooldowns when running a tag.
    The parameter requires 2 values to be passed: ``rate`` and ``per`` integers.
    The ``rate`` is the number of times the tag can be used every ``per`` seconds.

    The payload requires a ``key`` value, which is the key used to store the cooldown.
    A key should be any string that is unique. If a channel's ID is passed as a key,
    the tag's cooldown will be enforced on that channel. Running the tag in a separate channel
    would have a different cooldown with the same ``rate`` and ``per`` values.

    The payload also has an optional ``message`` value, which is the message to be sent when the
    cooldown is exceeded. If no message is passed, the default message will be sent instead.
    The cooldown message supports 2 blocks: ``key`` and ``retry_after``.

    **Usage:** ``{cooldown(<rate>|<per>):<key>|[message]}``

    **Payload:** key, message

    **Parameter:** rate, per

    **Examples:**

    .. tagscript::

        {cooldown(1|10):{author(id)}}
        the tag author used the tag more than once in 10 seconds
        The bucket for 741074175875088424 has reached its cooldown. Retry in 3.25 seconds."

        {cooldown(3|3):{channel(id)}|Slow down! This tag can only be used 3 times per 3 seconds per channel. Try again in **{retry_after}** seconds."}
        the tag was used more than 3 times in 3 seconds in a channel
        Slow down! This tag can only be used 3 times per 3 seconds per channel. Try again in **0.74** seconds.
    """

    ACCEPTED_NAMES = ("cooldown",)
    COOLDOWNS: Dict[Any, CooldownMapping] = {}

    @classmethod
    def create_cooldown(cls, key: Any, rate: int, per: int) -> CooldownMapping:
        cooldown = CooldownMapping.from_cooldown(rate, per, lambda x: x)
        cls.COOLDOWNS[key] = cooldown
        return cooldown

    def process(self, ctx: Context) -> Optional[str]:
        verb = ctx.verb
        try:
            rate, per = helper_split(verb.parameter, maxsplit=1)
            per = int(per)
            rate = float(rate)
        except (ValueError, TypeError):
            return

        if split := helper_split(verb.payload, False, maxsplit=1):
            key, message = split
        else:
            key = verb.payload
            message = None

        cooldown_key = ctx.response.extra_kwargs.get("cooldown_key")
        if cooldown_key is None:
            cooldown_key = ctx.original_message
        try:
            cooldown = self.COOLDOWNS[cooldown_key]
            base = cooldown._cooldown
            if (rate, per) != (base.rate, base.per):
                cooldown = self.create_cooldown(cooldown_key, rate, per)
        except KeyError:
            cooldown = self.create_cooldown(cooldown_key, rate, per)

        current = time.time()
        bucket = cooldown.get_bucket(key, current)
        retry_after = bucket.update_rate_limit(current)
        if retry_after:
            retry_after = round(retry_after, 2)
            if message:
                message = message.replace("{key}", str(key)).replace(
                    "{retry_after}", str(retry_after)
                )
            else:
                message = f"The bucket for {key} has reached its cooldown. Retry in {retry_after} seconds."
            raise CooldownExceeded(message, bucket, key, retry_after)
        return ""



import re
from inspect import isawaitable
from typing import Any, Awaitable, Callable, Union

__all__ = ("escape_content", "maybe_await")

pattern = re.compile(r"(?<!\\)([{():|}])")


def _sub_match(match: re.Match) -> str:
    return "\\" + match.group(1)


def escape_content(string: str) -> str:
    """
    Escapes given input to avoid tampering with engine/block behavior.

    Returns
    -------
    str
        The escaped content.
    """
    if string is None:
        return
    return pattern.sub(_sub_match, string)


async def maybe_await(func: Union[Callable[..., Any], Awaitable[Any]], *args, **kwargs) -> Any:
    """
    Await the given function if it is awaitable or call it synchronously.
    Returns
    -------
    Any
        The result of the awaitable function.
    """
    value = func(*args, **kwargs)
    return await value if isawaitable(value) else value



import random
from typing import Optional

from ..interface import verb_required_block
from ..interpreter import Context


class RandomBlock(verb_required_block(True, payload=True)):
    """
    Pick a random item from a list of strings, split by either ``~``
    or ``,``. An optional seed can be provided to the parameter to
    always choose the same item when using that seed.

    **Usage:** ``{random([seed]):<list>}``

    **Aliases:** ``#, rand``

    **Payload:** list

    **Parameter:** seed, None

    **Examples:**

    .. tagscript::

        {random:Carl,Harold,Josh} attempts to pick the lock!
        Possible Outputs:
        Josh attempts to pick the lock!
        Carl attempts to pick the lock!
        Harold attempts to pick the lock!

        {=(insults):You're so ugly that you went to the salon and it took 3 hours just to get an estimate.~I'll never forget the first time we met, although I'll keep trying.~You look like a before picture.}
        {=(insult):{#:{insults}}}
        {insult}
        Assigns a random insult to the insult variable
    """

    ACCEPTED_NAMES = ("random", "#", "rand")

    def process(self, ctx: Context) -> Optional[str]:
        spl = []
        if "~" in ctx.verb.payload:
            spl = ctx.verb.payload.split("~")
        else:
            spl = ctx.verb.payload.split(",")
        random.seed(ctx.verb.parameter)

        return random.choice(spl)



import re
from typing import List, Optional

__all__ = ("implicit_bool", "helper_parse_if", "helper_split", "helper_parse_list_if")

SPLIT_REGEX = re.compile(r"(?<!\\)\|")
BOOL_LOOKUP = {"true": True, "false": False}  # potentially add more bool values


def implicit_bool(string: str) -> Optional[bool]:
    """
    Parse a string to a boolean.

    >>> implicit_bool("true")
    True
    >>> implicit_bool("FALSE")
    False
    >>> implicit_bool("abc")
    None

    Parameters
    ----------
    string: str
        The string to convert.

    Returns
    -------
    bool
        The boolean value of the string.
    None
        The string failed to parse.
    """
    return BOOL_LOOKUP.get(string.lower())


def helper_parse_if(string: str) -> Optional[bool]:
    """
    Parse an expression string to a boolean.

    >>> helper_parse_if("this == this")
    True
    >>> helper_parse_if("2>3")
    False
    >>> helper_parse_if("40 >= 40")
    True
    >>> helper_parse_if("False")
    False
    >>> helper_parse_if("1")
    None

    Parameters
    ----------
    string: str
        The string to convert.

    Returns
    -------
    bool
        The boolean value of the expression.
    None
        The expression failed to parse.
    """
    value = implicit_bool(string)
    if value is not None:
        return value
    try:
        if "!=" in string:
            spl = string.split("!=")
            return spl[0].strip() != spl[1].strip()
        if "==" in string:
            spl = string.split("==")
            return spl[0].strip() == spl[1].strip()
        if ">=" in string:
            spl = string.split(">=")
            return float(spl[0].strip()) >= float(spl[1].strip())
        if "<=" in string:
            spl = string.split("<=")
            return float(spl[0].strip()) <= float(spl[1].strip())
        if ">" in string:
            spl = string.split(">")
            return float(spl[0].strip()) > float(spl[1].strip())
        if "<" in string:
            spl = string.split("<")
            return float(spl[0].strip()) < float(spl[1].strip())
    except:
        pass


def helper_split(
    split_string: str, easy: bool = True, *, maxsplit: int = None
) -> Optional[List[str]]:
    """
    A helper method to universalize the splitting logic used in multiple
    blocks and adapters. Please use this wherever a verb needs content to
    be chopped at | , or ~!

    >>> helper_split("this, should|work")
    ["this, should", "work"]
    """
    args = (maxsplit,) if maxsplit is not None else ()
    if "|" in split_string:
        return SPLIT_REGEX.split(split_string, *args)
    if easy and "~" in split_string:
        return split_string.split("~", *args)
    if easy and "," in split_string:
        return split_string.split(",", *args)
    return


def helper_parse_list_if(if_string):
    split = helper_split(if_string, False)
    if split is None:
        return [helper_parse_if(if_string)]
    return [helper_parse_if(item) for item in split]



from typing import Callable

from ..interface import Adapter
from ..verb import Verb


class FunctionAdapter(Adapter):
    __slots__ = ("fn",)

    def __init__(self, function_pointer: Callable[[], str]):
        self.fn = function_pointer
        super().__init__()

    def __repr__(self):
        return f"<{type(self).__qualname__} fn={self.fn!r}>"

    def get_value(self, ctx: Verb) -> str:
        return str(self.fn())



from datetime import datetime, timezone
from typing import Optional

from ..interface import Block
from ..interpreter import Context


class StrfBlock(Block):
    """
    The strf block converts and formats timestamps based on `strftime formatting spec <https://strftime.org/>`_.
    Two types of timestamps are supported: ISO and epoch.
    If a timestamp isn't passed, the current UTC time is used.

    Invoking this block with `unix` will return the current Unix timestamp.

    **Usage:** ``{strf([timestamp]):<format>}``

    **Aliases:** ``unix``

    **Payload:** format, None

    **Parameter:** timestamp

    **Example:**

    .. tagscript::

        {strf:%Y-%m-%d}
        2021-07-11

        {strf({user(timestamp)}):%c}
        Fri Jun 29 21:10:28 2018

        {strf(1420070400):%A %d, %B %Y}
        Thursday 01, January 2015

        {strf(2019-10-09T01:45:00.805000):%H:%M %d-%B-%Y}
        01:45 09-October-2019

        {unix}
        1629182008
    """

    ACCEPTED_NAMES = ("strf", "unix")

    def process(self, ctx: Context) -> Optional[str]:
        if ctx.verb.declaration.lower() == "unix":
            return str(int(datetime.now(timezone.utc).timestamp()))
        if not ctx.verb.payload:
            return None
        if ctx.verb.parameter:
            if ctx.verb.parameter.isdigit():
                try:
                    t = datetime.fromtimestamp(int(ctx.verb.parameter))
                except:
                    return
            else:
                try:
                    t = datetime.fromisoformat(ctx.verb.parameter)
                    # converts datetime.__str__ to datetime
                except ValueError:
                    return
        else:
            t = datetime.now()
        if not t.tzinfo:
            t = t.replace(tzinfo=timezone.utc)
        return t.strftime(ctx.verb.payload)



from typing import Optional

from ..interface import Block, verb_required_block
from ..interpreter import Context


class CommandBlock(verb_required_block(True, payload=True)):
    """
    Run a command as if the tag invoker had ran it. Only 3 command
    blocks can be used in a tag.

    **Usage:** ``{command:<command>}``

    **Aliases:** ``c, com, command``

    **Payload:** command

    **Parameter:** None

    **Examples:**

    .. tagscript::

        {c:ping}
        # invokes ping command

        {c:ban {target(id)} Chatflood/spam}
        # invokes ban command on the pinged user with the reason as "Chatflood/spam"
    """

    ACCEPTED_NAMES = ("c", "com", "command")

    def __init__(self, limit: int = 3):
        self.limit = limit
        super().__init__()

    def process(self, ctx: Context) -> Optional[str]:
        command = ctx.verb.payload.strip()
        actions = ctx.response.actions.get("commands")
        if actions:
            if len(actions) >= self.limit:
                return f"`COMMAND LIMIT REACHED ({self.limit})`"
        else:
            ctx.response.actions["commands"] = []
        ctx.response.actions["commands"].append(command)
        return ""


class OverrideBlock(Block):
    """
    Override a command's permission requirements. This can override
    mod, admin, or general user permission requirements when running commands
    with the :ref:`Command Block`. Passing no parameter will default to overriding
    all permissions.

    In order to add a tag with the override block, the tag author must have ``Manage
    Server`` permissions.

    This will not override bot owner commands or command checks.

    **Usage:** ``{override(["admin"|"mod"|"permissions"]):[command]}``

    **Payload:** command

    **Parameter:** "admin", "mod", "permissions"

    **Examples:**
    
    .. tagscript::

        {override}
        overrides all commands and permissions

        {override(admin)}
        overrides commands that require the admin role

        {override(permissions)}
        {override(mod)}
        overrides commands that require the mod role or have user permission requirements
    """

    ACCEPTED_NAMES = ("override",)

    def process(self, ctx: Context) -> Optional[str]:
        param = ctx.verb.parameter
        if not param:
            ctx.response.actions["overrides"] = {"admin": True, "mod": True, "permissions": True}
            return ""

        param = param.strip().lower()
        if param not in ("admin", "mod", "permissions"):
            return None
        overrides = ctx.response.actions.get(
            "overrides", {"admin": False, "mod": False, "permissions": False}
        )
        overrides[param] = True
        ctx.response.actions["overrides"] = overrides
        return ""



from inspect import ismethod

from ..interface import Adapter
from ..verb import Verb


class SafeObjectAdapter(Adapter):
    __slots__ = ("object",)

    def __init__(self, base):
        self.object = base

    def __repr__(self):
        return f"<{type(self).__qualname__} object={repr(self.object)}>"

    def get_value(self, ctx: Verb) -> str:
        if ctx.parameter is None:
            return str(self.object)
        if ctx.parameter.startswith("_") or "." in ctx.parameter:
            return
        try:
            attribute = getattr(self.object, ctx.parameter)
        except AttributeError:
            return
        if ismethod(attribute):
            return
        if isinstance(attribute, float):
            attribute = int(attribute)
        return str(attribute)



from typing import Optional

from ..interface import verb_required_block
from ..interpreter import Context


class RedirectBlock(verb_required_block(True, parameter=True)):
    """
    Redirects the tag response to either the given channel, the author's DMs,
    or uses a reply based on what is passed to the parameter.

    **Usage:** ``{redirect(<"dm"|"reply"|channel>)}``

    **Payload:** None

    **Parameter:** "dm", "reply", channel

    **Examples:**
    
    .. tagscript::

        {redirect(dm)}
        {redirect(reply)}
        {redirect(#general)}
        {redirect(626861902521434160)}
    """

    ACCEPTED_NAMES = ("redirect",)

    def process(self, ctx: Context) -> Optional[str]:
        param = ctx.verb.parameter.strip()
        if param.lower() == "dm":
            target = "dm"
        elif param.lower() == "reply":
            target = "reply"
        else:
            target = param
        ctx.response.actions["target"] = target
        return ""



from urllib.parse import unquote, unquote_plus

from ..interface import verb_required_block
from ..interpreter import Context


class URLDecodeBlock(verb_required_block(True, payload=True)):
    """
    This block will decode a given url into a string
    with non-url compliant characters replaced. Using ``+`` as the parameter
    will replace spaces with ``+`` rather than ``%20``.

    **Usage:** ``{urldecode(["+"]):<string>}``

    **Payload:** string

    **Parameter:** "+", None

    **Examples:**

    .. tagscript::

        {urldecode:covid-19%20sucks}
        # covid-19 sucks

        {urldecode(+):im+stuck+at+home+writing+docs}
        # im stuck at home writing docs

    This block is just the reverse of the urlencode block
    """

    ACCEPTED_NAMES = ("urldecode",)

    def process(self, ctx: Context):
        method = unquote_plus if ctx.verb.parameter == "+" else unquote
        return method(ctx.verb.payload)



import random
from typing import Optional

from ..interface import verb_required_block
from ..interpreter import Context


class RangeBlock(verb_required_block(True, payload=True)):
    """
    The range block picks a random number from a range of numbers seperated by ``-``.
    The number range is inclusive, so it can pick the starting/ending number as well.
    Using the rangef block will pick a number to the tenth decimal place.

    An optional seed can be provided to the parameter to always choose the same item when using that seed.

    **Usage:** ``{range([seed]):<lowest-highest>}``

    **Aliases:** ``rangef``

    **Payload:** number

    **Parameter:** seed, None

    **Examples:**

    .. tagscript::

        Your lucky number is {range:10-30}!
        Your lucky number is 14!
        Your lucky number is 25!

        {=(height):{rangef:5-7}}
        I am guessing your height is {height}ft.
        I am guessing your height is 5.3ft.
    """

    ACCEPTED_NAMES = ("rangef", "range")

    def process(self, ctx: Context) -> Optional[str]:
        try:
            spl = ctx.verb.payload.split("-")
            random.seed(ctx.verb.parameter)
            if ctx.verb.declaration.lower() == "rangef":
                lower = float(spl[0])
                upper = float(spl[1])
                base = random.randint(lower * 10, upper * 10) / 10
                return str(base)
                # base = random.randint(lower, upper)
                # if base == upper:
                #     return str(base)
                # if ctx.verb.parameter != None:
                #     random.seed(ctx.verb.parameter+"float")
                # else:
                #     random.seed(None)
                # return str(str(base)+"."+str(random.randint(1,9)))
            else:
                lower = int(float(spl[0]))
                upper = int(float(spl[1]))
                return str(random.randint(lower, upper))
        except:
            return None



import random
from typing import Optional

from ..interface import verb_required_block
from ..interpreter import Context


class FiftyFiftyBlock(verb_required_block(True, payload=True)):
    """
    The fifty-fifty block has a 50% change of returning the payload, and 50% chance of returning null.

    **Usage:**  ``{50:<message>}``

    **Aliases:**  ``5050, ?``

    **Payload:**  message

    **Parameter:**  None

    **Examples:**

    .. tagscript::

        I pick {if({5050:.}!=):heads|tails}
        # I pick heads
    """

    ACCEPTED_NAMES = ("5050", "50", "?")

    def process(self, ctx: Context) -> Optional[str]:
        return random.choice(["", ctx.verb.payload])



# isort: off
from .helpers import *

# isort: on
from .assign import AssignmentBlock
from .breakblock import BreakBlock
from .command import CommandBlock, OverrideBlock
from .control import AllBlock, AnyBlock, IfBlock
from .cooldown import CooldownBlock
from .embedblock import EmbedBlock
from .fiftyfifty import FiftyFiftyBlock
from .loosevariablegetter import LooseVariableGetterBlock
from .mathblock import MathBlock
from .randomblock import RandomBlock
from .range import RangeBlock
from .redirect import RedirectBlock
from .replaceblock import PythonBlock, ReplaceBlock
from .require_blacklist import BlacklistBlock, RequireBlock
from .shortcutredirect import ShortCutRedirectBlock
from .stopblock import StopBlock
from .strf import StrfBlock
from .strictvariablegetter import StrictVariableGetterBlock
from .substr import SubstringBlock
from .urlencodeblock import URLEncodeBlock
from .urldecodeblock import URLDecodeBlock

__all__ = (
    "implicit_bool",
    "helper_parse_if",
    "helper_parse_list_if",
    "helper_split",
    "AllBlock",
    "AnyBlock",
    "AssignmentBlock",
    "BlacklistBlock",
    "BreakBlock",
    "CommandBlock",
    "CooldownBlock",
    "EmbedBlock",
    "FiftyFiftyBlock",
    "IfBlock",
    "LooseVariableGetterBlock",
    "MathBlock",
    "OverrideBlock",
    "PythonBlock",
    "RandomBlock",
    "RangeBlock",
    "RedirectBlock",
    "ReplaceBlock",
    "RequireBlock",
    "ShortCutRedirectBlock",
    "StopBlock",
    "StrfBlock",
    "StrictVariableGetterBlock",
    "SubstringBlock",
    "URLEncodeBlock",
    "URLDecodeBlock"
)



from typing import Optional

from ..interface import verb_required_block
from ..interpreter import Context
from . import helper_parse_if, helper_parse_list_if, helper_split


def parse_into_output(payload: str, result: Optional[bool]) -> Optional[str]:
    if result is None:
        return
    try:
        output = helper_split(payload, False)
        if output != None and len(output) == 2:
            if result:
                return output[0]
            else:
                return output[1]
        elif result:
            return payload
        else:
            return ""
    except:
        return


ImplicitPPRBlock = verb_required_block(True, payload=True, parameter=True)


class AnyBlock(ImplicitPPRBlock):
    """
    The any block checks that any of the passed expressions are true.
    Multiple expressions can be passed to the parameter by splitting them with ``|``.

    The payload is a required message that must be split by ``|``.
    If the expression evaluates true, then the message before the ``|`` is returned, else the message after is returned.

    **Usage:** ``{any(<expression|expression|...>):<message>}``

    **Aliases:** ``or``

    **Payload:** message

    **Parameter:** expression
    
    **Examples:**

    .. tagscript::

        {any({args}==hi|{args}==hello|{args}==heyy):Hello {user}!|How rude.}
        if {args} is hi
        Hello sravan#0001!

        if {args} is what's up!
        How rude.
    """

    ACCEPTED_NAMES = ("any", "or")

    def process(self, ctx: Context) -> Optional[str]:
        result = any(helper_parse_list_if(ctx.verb.parameter) or [])
        return parse_into_output(ctx.verb.payload, result)


class AllBlock(ImplicitPPRBlock):
    """
    The all block checks that all of the passed expressions are true.
    Multiple expressions can be passed to the parameter by splitting them with ``|``.

    The payload is a required message that must be split by ``|``.
    If the expression evaluates true, then the message before the ``|`` is returned, else the message after is returned.

    **Usage:** ``{all(<expression|expression|...>):<message>}``

    **Aliases:** ``and``

    **Payload:** message

    **Parameter:** expression

    **Examples:**

    .. tagscript::

        {all({args}>=100|{args}<=1000):You picked {args}.|You must provide a number between 100 and 1000.}
        # if {args} is 52
        You must provide a number between 100 and 1000.

        # if {args} is 282
        You picked 282.
    """

    ACCEPTED_NAMES = ("all", "and")

    def process(self, ctx: Context) -> Optional[str]:
        result = all(helper_parse_list_if(ctx.verb.parameter) or [])
        return parse_into_output(ctx.verb.payload, result)


class IfBlock(ImplicitPPRBlock):
    """
    The if block returns a message based on the passed expression to the parameter.
    An expression is represented by two values compared with an operator.

    The payload is a required message that must be split by ``|``.
    If the expression evaluates true, then the message before the ``|`` is returned, else the message after is returned.

    **Expression Operators:**

    +----------+--------------------------+---------+---------------------------------------------+
    | Operator | Check                    | Example | Description                                 |
    +==========+==========================+=========+=============================================+
    | ``==``   | equality                 | a==a    | value 1 is equal to value 2                 |
    +----------+--------------------------+---------+---------------------------------------------+
    | ``!=``   | inequality               | a!=b    | value 1 is not equal to value 2             |
    +----------+--------------------------+---------+---------------------------------------------+
    | ``>``    | greater than             | 5>3     | value 1 is greater than value 2             |
    +----------+--------------------------+---------+---------------------------------------------+
    | ``<``    | less than                | 4<8     | value 1 is less than value 2                |
    +----------+--------------------------+---------+---------------------------------------------+
    | ``>=``   | greater than or equality | 10>=10  | value 1 is greater than or equal to value 2 |
    +----------+--------------------------+---------+---------------------------------------------+
    | ``<=``   | less than or equality    | 5<=6    | value 1 is less than or equal to value 2    |
    +----------+--------------------------+---------+---------------------------------------------+

    **Usage:** ``{if(<expression>):<message>]}``

    **Payload:** message

    **Parameter:** expression

    **Examples:**
    
    .. tagscript::

        {if({args}==63):You guessed it! The number I was thinking of was 63!|Too {if({args}<63):low|high}, try again.}
        if args is 63
        You guessed it! The number I was thinking of was 63!

        if args is 73
        Too low, try again.

        if args is 14
        Too high, try again.
    """

    ACCEPTED_NAMES = ("if",)

    def process(self, ctx: Context) -> Optional[str]:
        result = helper_parse_if(ctx.verb.parameter)
        return parse_into_output(ctx.verb.payload, result)



from urllib.parse import quote, quote_plus

from ..interface import verb_required_block
from ..interpreter import Context


class URLEncodeBlock(verb_required_block(True, payload=True)):
    """
    This block will encode a given string into a properly formatted url
    with non-url compliant characters replaced. Using ``+`` as the parameter
    will replace spaces with ``+`` rather than ``%20``.

    **Usage:** ``{urlencode(["+"]):<string>}``

    **Payload:** string

    **Parameter:** "+", None

    **Example:**
    
    .. tagscript::

        {urlencode:covid-19 sucks}
        covid-19%20sucks

        {urlencode(+):im stuck at home writing docs}
        im+stuck+at+home+writing+docs

        You can use this to search up blocks
        Eg if {args} is command block

        <https://btagscript.readthedocs.io/en/latest/search.html?q={urlencode(+):{args}}&check_keywords=yes&area=default>
        <https://btagscript.readthedocs.io/en/latest/search.html?q=command+block&check_keywords=yes&area=default>
    """

    ACCEPTED_NAMES = ("urlencode",)

    def process(self, ctx: Context):
        method = quote_plus if ctx.verb.parameter == "+" else quote
        return method(ctx.verb.payload)



from __future__ import annotations

from functools import lru_cache
from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from ..interpreter import Context


__all__ = ("Block", "verb_required_block")


class Block:
    """
    The base class for TagScript blocks.

    Implementations must subclass this to create new blocks.

    Attributes
    ----------
    ACCEPTED_NAMES: Tuple[str, ...]
        The accepted names for this block. This ideally should be set as a class attribute.
    """

    ACCEPTED_NAMES = ()

    def __repr__(self):
        return f"<{type(self).__qualname__} at {hex(id(self))}>"

    @classmethod
    def will_accept(cls, ctx: Context) -> bool:
        """
        Describes whether the block is valid for the given :class:`~bTagScript.interpreter.Context`.

        Parameters
        ----------
        ctx: Context
            The context object containing the TagScript :class:`~bTagScript.verb.Verb`.

        Returns
        -------
        bool
            Whether the block should be processed for this :class:`~bTagScript.interpreter.Context`.
        """
        dec = ctx.verb.declaration.lower()
        return dec in cls.ACCEPTED_NAMES

    def pre_process(self, ctx: Context):
        return None

    def process(self, ctx: Context) -> Optional[str]:
        """
        Processes the block's actions for a given :class:`~bTagScript.interpreter.Context`.

        Subclasses must implement this.

        Parameters
        ----------
        ctx: Context
            The context object containing the TagScript :class:`~bTagScript.verb.Verb`.

        Returns
        -------
        Optional[str]
            The block's processed value.

        Raises
        ------
        NotImplementedError
            The subclass did not implement this required method.
        """
        raise NotImplementedError

    def post_process(self, ctx: Context):
        return None


@lru_cache(maxsize=None)
def verb_required_block(
    implicit: bool,
    *,
    parameter: bool = False,
    payload: bool = False,
) -> Block:
    """
    Get a Block subclass that requires a verb to implicitly or explicitly have a parameter or payload passed.

    Parameters
    ----------
    implicit: bool
        Specifies whether the value is required to be passed implicitly or explicitly.
        ``{block()}`` would be allowed if implicit is False.
    parameter: bool
        Passing True will cause the block to require a parameter to be passed.
    payload: bool
        Passing True will cause the block to require the payload to be passed.
    """
    check = (lambda x: x) if implicit else (lambda x: x is not None)

    class RequireMeta(type):
        def __repr__(self):
            return f"VerbRequiredBlock(implicit={implicit!r}, payload={payload!r}, parameter={parameter!r})"

    class VerbRequiredBlock(Block, metaclass=RequireMeta):
        def will_accept(self, ctx: Context) -> bool:
            verb = ctx.verb
            if payload and not check(verb.payload):
                return False
            if parameter and not check(verb.parameter):
                return False
            return super().will_accept(ctx)

    return VerbRequiredBlock



from .discordadapters import *
from .functionadapter import FunctionAdapter
from .intadapter import IntAdapter
from .objectadapter import SafeObjectAdapter
from .stringadapter import StringAdapter

__all__ = (
    "SafeObjectAdapter",
    "StringAdapter",
    "IntAdapter",
    "FunctionAdapter",
    "AttributeAdapter",
    "MemberAdapter",
    "ChannelAdapter",
    "GuildAdapter",
)



from typing import Optional

from ..interface import verb_required_block
from ..interpreter import Context


class RequireBlock(verb_required_block(True, parameter=True)):
    """
    The require block will attempt to convert the given parameter into a channel
    or role, using name or ID. If the user running the tag is not in the targeted
    channel or doesn't have the targeted role, the tag will stop processing and
    it will send the response if one is given. Multiple role or channel
    requirements can be given, and should be split by a ",".

    **Usage:** ``{require(<role,channel>):[response]}``

    **Aliases:** ``whitelist``

    **Payload:** response, None

    **Parameter:** role, channel

    **Examples:**
    
    .. tagscript::

        {require(Moderator)}
        {require(#general, #bot-cmds):This tag can only be run in #general and #bot-cmds.}
        {require(757425366209134764, 668713062186090506, 737961895356792882):You aren't allowed to use this tag.}
    """

    ACCEPTED_NAMES = ("require", "whitelist")

    def process(self, ctx: Context) -> Optional[str]:
        actions = ctx.response.actions.get("requires")
        if actions:
            return None
        ctx.response.actions["requires"] = {
            "items": [i.strip() for i in ctx.verb.parameter.split(",")],
            "response": ctx.verb.payload,
        }
        return ""


class BlacklistBlock(verb_required_block(True, parameter=True)):
    """
    The blacklist block will attempt to convert the given parameter into a channel
    or role, using name or ID. If the user running the tag is in the targeted
    channel or has the targeted role, the tag will stop processing and
    it will send the response if one is given. Multiple role or channel
    requirements can be given, and should be split by a ",".

    **Usage:** ``{blacklist(<role,channel>):[response]}``

    **Payload:** response, None

    **Parameter:** role, channel

    **Examples:**

    .. tagscript:: 
        {blacklist(Muted)}
        {blacklist(#support):This tag is not allowed in #support.}
        {blacklist(Tag Blacklist, 668713062186090506):You are blacklisted from using tags.}
    """

    ACCEPTED_NAMES = ("blacklist",)

    def process(self, ctx: Context) -> Optional[str]:
        actions = ctx.response.actions.get("blacklist")
        if actions:
            return None
        ctx.response.actions["blacklist"] = {
            "items": [i.strip() for i in ctx.verb.parameter.split(",")],
            "response": ctx.verb.payload,
        }
        return ""



from __future__ import annotations

import logging
from itertools import islice
from typing import Any, Dict, List, Optional, Tuple

from .exceptions import ProcessError, StopError, TagScriptError, WorkloadExceededError
from .interface import Adapter, Block
from .utils import maybe_await
from .verb import Verb

__all__ = (
    "Interpreter",
    "AsyncInterpreter",
    "Context",
    "Response",
    "Node",
    "build_node_tree",
)

log = logging.getLogger(__name__)

AdapterDict = Dict[str, Adapter]


class Node:
    """
    A low-level object representing a bracketed block.
    Attributes
    ----------
    coordinates: Tuple[int, int]
        The start and end position of the bracketed text block.
    verb: Optional[Verb]
        The determined Verb for this node.
    output:
        The `Block` processed output for this node.
    """

    __slots__ = ("output", "verb", "coordinates")

    def __init__(self, coordinates: Tuple[int, int], verb: Optional[Verb] = None):
        self.output: Optional[str] = None
        self.verb = verb
        self.coordinates = coordinates

    def __str__(self):
        return str(self.verb) + " at " + str(self.coordinates)

    def __repr__(self):
        return f"<Node verb={self.verb!r} coordinates={self.coordinates!r} output={self.output!r}>"


def build_node_tree(message: str) -> List[Node]:
    """
    Function that finds all possible nodes in a string.
    Returns
    -------
    List[Node]
        A list of all possible text bracket blocks.
    """
    nodes = []
    previous = r""

    starts = []
    for i, ch in enumerate(message):
        if ch == "{" and previous != r"\\":
            starts.append(i)
        if ch == "}" and previous != r"\\":
            if not starts:
                continue
            coords = (starts.pop(), i)
            n = Node(coords)
            nodes.append(n)

        previous = ch
    return nodes


class Response:
    """
    An object containing information on a completed TagScript process.
    Attributes
    ----------
    body: str
        The cleaned message with all verbs interpreted.
    actions: Dict[str, Any]
        A dictionary that blocks can access and modify to define post-processing actions.
    variables: Dict[str, Adapter]
        A dictionary of variables that blocks such as the `LooseVariableGetterBlock` can access.
    extra_kwargs: Dict[str, Any]
        A dictionary of extra keyword arguments that blocks can use to define their own behavior.
    """

    __slots__ = ("body", "actions", "variables", "extra_kwargs")

    def __init__(self, *, variables: AdapterDict = None, extra_kwargs: Dict[str, Any] = None):
        self.body: str = None
        self.actions: Dict[str, Any] = {}
        self.variables: AdapterDict = variables if variables is not None else {}
        self.extra_kwargs: Dict[str, Any] = extra_kwargs if extra_kwargs is not None else {}

    def __repr__(self):
        return (
            f"<Response body={self.body!r} actions={self.actions!r} variables={self.variables!r}>"
        )


class Context:
    """
    An object containing data on the TagScript block processed by the interpreter.
    This class is passed to adapters and blocks during processing.
    Attributes
    ----------
    verb: Verb
        The Verb object representing a TagScript block.
    original_message: str
        The original message passed to the interpreter.
    interpreter: Interpreter
        The interpreter processing the TagScript.
    """

    __slots__ = ("verb", "original_message", "interpreter", "response")

    def __init__(self, verb: Verb, res: Response, interpreter: Interpreter, og: str):
        self.verb: Verb = verb
        self.original_message: str = og
        self.interpreter: Interpreter = interpreter
        self.response: Response = res

    def __repr__(self):
        return f"<Context verb={self.verb!r}>"


class Interpreter:
    """
    The TagScript interpreter.
    Attributes
    ----------
    blocks: List[Block]
        A list of blocks to be used for TagScript processing.
    """

    __slots__ = ("blocks",)

    def __init__(self, blocks: List[Block]):
        self.blocks: List[Block] = blocks

    def __repr__(self):
        return f"<{type(self).__name__} blocks={self.blocks!r}>"

    def _get_context(
        self,
        node: Node,
        final: str,
        *,
        response: Response,
        original_message: str,
        verb_limit: int,
        dot_parameter: bool,
    ) -> Context:
        # Get the updated verb string from coordinates and make the context
        start, end = node.coordinates
        node.verb = Verb(final[start : end + 1], limit=verb_limit, dot_parameter=dot_parameter)
        return Context(node.verb, response, self, original_message)

    def _get_acceptors(self, ctx: Context) -> List[Block]:
        acceptors = [b for b in self.blocks if b.will_accept(ctx)]
        log.debug("%r acceptors: %r", ctx, acceptors)
        return acceptors

    def _process_blocks(self, ctx: Context, node: Node) -> Optional[str]:
        acceptors = self._get_acceptors(ctx)
        for b in acceptors:
            value = b.process(ctx)
            if value is not None:  # Value found? We're done here.
                value = str(value)
                node.output = value
                return value

    @staticmethod
    def _check_workload(charlimit: int, total_work: int, output: str) -> Optional[int]:
        if not charlimit:
            return
        total_work += len(output)
        if total_work > charlimit:
            raise WorkloadExceededError(
                "The TSE interpreter had its workload exceeded. The total characters "
                f"attempted were {total_work}/{charlimit}"
            )
        return total_work

    @staticmethod
    def _text_deform(start: int, end: int, final: str, output: str) -> Tuple[str, int]:
        message_slice_len = (end + 1) - start
        replacement_len = len(output)
        differential = (
            replacement_len - message_slice_len
        )  # The change in size of `final` after the change is applied
        final = final[:start] + output + final[end + 1 :]
        return final, differential

    @staticmethod
    def _translate_nodes(node_ordered_list: List[Node], index: int, start: int, differential: int):
        for future_n in islice(node_ordered_list, index + 1, None):
            new_start = None
            new_end = None
            if future_n.coordinates[0] > start:
                new_start = future_n.coordinates[0] + differential
            else:
                new_start = future_n.coordinates[0]

            if future_n.coordinates[1] > start:
                new_end = future_n.coordinates[1] + differential
            else:
                new_end = future_n.coordinates[1]
            future_n.coordinates = (new_start, new_end)

    def _solve(
        self,
        message: str,
        node_ordered_list: List[Node],
        response: Response,
        *,
        charlimit: int,
        verb_limit: int = 2000,
        dot_parameter: bool,
    ):
        final = message
        total_work = 0
        for index, node in enumerate(node_ordered_list):
            start, end = node.coordinates
            ctx = self._get_context(
                node,
                final,
                response=response,
                original_message=message,
                verb_limit=verb_limit,
                dot_parameter=dot_parameter,
            )
            log.debug("Processing context %r at (%r, %r)", ctx, start, end)
            try:
                output = self._process_blocks(ctx, node)
            except StopError as exc:
                log.debug("StopError raised on node %r", node, exc_info=exc)
                return final[:start] + exc.message
            if output is None:
                continue  # If there was no value output, no need to text deform.

            total_work = self._check_workload(charlimit, total_work, output)
            final, differential = self._text_deform(start, end, final, output)
            self._translate_nodes(node_ordered_list, index, start, differential)
        return final

    @staticmethod
    def _return_response(response: Response, output: str) -> Response:
        if response.body is None:
            response.body = output.strip()
        else:
            # Dont override an overridden response.
            response.body = response.body.strip()
        return response

    def process(
        self,
        message: str,
        seed_variables: AdapterDict = None,
        *,
        charlimit: Optional[int] = None,
        dot_parameter: bool = False,
        **kwargs,
    ) -> Response:
        """
        Processes a given TagScript string.
        Parameters
        ----------
        message: str
            A TagScript string to be processed.
        seed_variables: Dict[str, Adapter]
            A dictionary containing strings to adapters to provide context variables for processing.
        charlimit: int
            The maximum characters to process.
        dot_parameter: bool
            Whether the parameter should be followed after a "." or use the default of parantheses.
        kwargs: Dict[str, Any]
            Additional keyword arguments that may be used by blocks during processing.
        Returns
        -------
        Response
            A response object containing the processed body, actions and variables.
        Raises
        ------
        TagScriptError
            A block intentionally raised an exception, most likely due to invalid user input.
        WorkloadExceededError
            Signifies the interpreter reached the character limit, if one was provided.
        ProcessError
            An unexpected error occurred while processing blocks.
        """
        response = Response(variables=seed_variables, extra_kwargs=kwargs)
        node_ordered_list = build_node_tree(message)
        try:
            output = self._solve(
                message,
                node_ordered_list,
                response,
                charlimit=charlimit,
                dot_parameter=dot_parameter,
            )
        except TagScriptError:
            raise
        except Exception as error:
            raise ProcessError(error, response, self) from error
        return self._return_response(response, output)


class AsyncInterpreter(Interpreter):
    """
    An asynchronous subclass of `Interpreter` that allows blocks to implement asynchronous methods.
    Synchronous blocks are still supported.
    This subclass has no additional attributes from the `Interpreter` class.
    See `Interpreter` for full documentation.
    """

    async def _get_acceptors(self, ctx: Context) -> List[Block]:
        return [b for b in self.blocks if await maybe_await(b.will_accept, ctx)]

    async def _process_blocks(self, ctx: Context, node: Node) -> Optional[str]:
        acceptors = await self._get_acceptors(ctx)
        for b in acceptors:
            value = await maybe_await(b.process, ctx)
            if value is not None:  # Value found? We're done here.
                value = str(value)
                node.output = value
                return value

    async def _solve(
        self,
        message: str,
        node_ordered_list: List[Node],
        response: Response,
        *,
        charlimit: int,
        verb_limit: int = 2000,
        dot_parameter: bool,
    ):
        final = message
        total_work = 0

        for index, node in enumerate(node_ordered_list):
            start, end = node.coordinates
            ctx = self._get_context(
                node,
                final,
                response=response,
                original_message=message,
                verb_limit=verb_limit,
                dot_parameter=dot_parameter,
            )
            try:
                output = await self._process_blocks(ctx, node)
            except StopError as exc:
                return final[:start] + exc.message
            if output is None:
                continue  # If there was no value output, no need to text deform.

            total_work = self._check_workload(charlimit, total_work, output)
            final, differential = self._text_deform(start, end, final, output)
            self._translate_nodes(node_ordered_list, index, start, differential)
        return final

    async def process(
        self,
        message: str,
        seed_variables: AdapterDict = None,
        *,
        charlimit: Optional[int] = None,
        dot_parameter: bool = False,
        **kwargs,
    ) -> Response:
        """
        Asynchronously process a given TagScript string.
        This method has no additional attributes from the `Interpreter` class.
        See `Interpreter.process` for full documentation.
        """
        response = Response(variables=seed_variables, extra_kwargs=kwargs)
        node_ordered_list = build_node_tree(message)
        try:
            output = await self._solve(
                message,
                node_ordered_list,
                response,
                charlimit=charlimit,
                dot_parameter=dot_parameter,
            )
        except TagScriptError:
            raise
        except Exception as error:
            raise ProcessError(error, response, self) from error
        return self._return_response(response, output)


from collections import namedtuple

from .adapter import *
from .block import *
from .exceptions import *
from .interface import *
from .interpreter import *
from .utils import *
from .verb import Verb

__version__ = "2.6.4"


# github co-pilot automatically wrote the documentation for this and I decided to keep it
class VersionInfo(namedtuple("VersionInfo", "major minor micro")):
    """
    Version information.

    Attributes
    ----------
    major: int
        Major version number.
    minor: int
        Minor version number.
    micro: int
        Micro version number.
    """

    __slots__ = ()

    def __str__(self):
        """
        Returns a string representation of the version information.

        Returns
        -------
        str
            String representation of the version information.
        """
        return "{major}.{minor}.{micro}".format(**self._asdict())

    @classmethod
    def from_str(cls, version):
        """
        Returns a VersionInfo instance from a string.

        Parameters
        ----------
        version: str
            String representation of the version information.

        Returns
        -------
        VersionInfo
            Version information.
        """
        return cls(*map(int, version.split(".")))


version_info = VersionInfo.from_str(__version__)



from ..interface import verb_required_block
from ..interpreter import Context


class ReplaceBlock(verb_required_block(True, payload=True, parameter=True)):
    """
    The replace block will replace specific characters in a string.
    The parameter should split by a ``,``, containing the characters to find
    before the command and the replacements after.

    **Usage:** ``{replace(<original,new>):<message>}``

    **Aliases:** ``sub``

    **Payload:** message

    **Parameter:** original, new

    .. tagscript::

        {replace(o,i):welcome to the server}
        welcime ti the server

        {replace(1,6):{args}}
        if {args} is 1637812
        6637862

        {replace(, ):Test}
        T e s t
    """

    ACCEPTED_NAMES = ("replace", "sub")

    def process(self, ctx: Context):
        try:
            before, after = ctx.verb.parameter.split(",", 1)
        except ValueError:
            return

        return ctx.verb.payload.replace(before, after)


class PythonBlock(verb_required_block(True, payload=True, parameter=True)):
    """
    The in block serves three different purposes depending on the alias that is used.

    The ``in`` alias checks if the parameter is anywhere in the payload.

    ``contain`` strictly checks if the parameter is the payload, split by whitespace.

    ``index`` finds the location of the parameter in the payload, split by whitespace.
    If the parameter string is not found in the payload, it returns 1.

    index is used to return the value of the string form the given list of

    **Usage:** ``{in(<string>):<payload>}``

    **Aliases:** ``index``, ``contains``

    **Payload:** payload

    **Parameter:** string

    **Examples:**
    
    .. tagscript::

        {in(apple pie):banana pie apple pie and other pie}
        # true
        {in(mute):How does it feel to be muted?}
        # true
        {in(a):How does it feel to be muted?}
        # false

        {contains(mute):How does it feel to be muted?}
        # false
        {contains(muted?):How does it feel to be muted?}
        # false

        {index(food):I love to eat food. everyone does.}
        # 4
        {index(pie):I love to eat food. everyone does.}
        # -1
    """

    def will_accept(self, ctx: Context):
        dec = ctx.verb.declaration.lower()
        return dec in ("contains", "in", "index")

    def process(self, ctx: Context):
        dec = ctx.verb.declaration.lower()
        if dec == "contains":
            return str(bool(ctx.verb.parameter in ctx.verb.payload.split())).lower()
        elif dec == "in":
            return str(bool(ctx.verb.parameter in ctx.verb.payload)).lower()
        else:
            try:
                return str(ctx.verb.payload.strip().split().index(ctx.verb.parameter))
            except ValueError:
                return "-1"



from ..interface import Adapter
from ..utils import escape_content
from ..verb import Verb


class StringAdapter(Adapter):
    __slots__ = ("string", "escape_content")

    def __init__(self, string: str, *, escape: bool = False):
        self.string: str = str(string)
        self.escape_content = escape

    def __repr__(self):
        return f"<{type(self).__qualname__} string={repr(self.string)}>"

    def get_value(self, ctx: Verb) -> str:
        return self.return_value(self.handle_ctx(ctx))

    def handle_ctx(self, ctx: Verb) -> str:
        if ctx.parameter is None:
            return self.string
        try:
            if "+" not in ctx.parameter:
                index = int(ctx.parameter) - 1
                splitter = " " if ctx.payload is None else ctx.payload
                return self.string.split(splitter)[index]
            else:
                index = int(ctx.parameter.replace("+", "")) - 1
                splitter = " " if ctx.payload is None else ctx.payload
                if ctx.parameter.startswith("+"):
                    return splitter.join(self.string.split(splitter)[: index + 1])
                elif ctx.parameter.endswith("+"):
                    return splitter.join(self.string.split(splitter)[index:])
                else:
                    return self.string.split(splitter)[index]
        except:
            return self.string

    def return_value(self, string: str) -> str:
        return escape_content(string) if self.escape_content else string



from typing import Optional

from ..interface import Block
from ..interpreter import Context
from . import helper_parse_if


class BreakBlock(Block):
    """
    The break block will force the tag output to only be the payload of this block, if the passed
    expresssion evaluates true.
    If no message is provided to the payload, the tag output will be empty.

    This differs from the `StopBlock` as the stop block stops all tagscript processing and returns
    its message while the break block continues to process blocks. If command blocks exist after
    the break block, they will still execute.

    **Usage:** ``{break(<expression>):[message]}``

    **Aliases:** ``short, shortcircuit``

    **Payload:**  message

    **Parameter:**  expression

    **Examples:**

    .. tagscript::
    
        {break({args}==):You did not provide any input.}
    """

    ACCEPTED_NAMES = ("break", "shortcircuit", "short")

    def process(self, ctx: Context) -> Optional[str]:
        if helper_parse_if(ctx.verb.parameter):
            ctx.response.body = ctx.verb.payload if ctx.verb.payload != None else ""
        return ""



from __future__ import division

import math
import operator
from typing import Optional

from pyparsing import (
    CaselessLiteral,
    Combine,
    Forward,
    Group,
    Literal,
    Optional,
    Word,
    ZeroOrMore,
    alphas,
    nums,
    oneOf,
)

from ..interface import Block
from ..interpreter import Context


class NumericStringParser(object):
    """
    Most of this code comes from the fourFn.py pyparsing example

    """

    def pushFirst(self, strg, loc, toks):
        self.exprStack.append(toks[0])

    def pushUMinus(self, strg, loc, toks):
        if toks and toks[0] == "-":
            self.exprStack.append("unary -")

    def __init__(self):
        """
        expop   :: '^'
        multop  :: '*' | '/'
        addop   :: '+' | '-'
        integer :: ['+' | '-'] '0'..'9'+
        atom    :: PI | E | real | fn '(' expr ')' | '(' expr ')'
        factor  :: atom [ expop factor ]*
        term    :: factor [ multop factor ]*
        expr    :: term [ addop term ]*
        """
        point = Literal(".")
        e = CaselessLiteral("E")
        fnumber = Combine(
            Word("+-" + nums, nums)
            + Optional(point + Optional(Word(nums)))
            + Optional(e + Word("+-" + nums, nums))
        )
        ident = Word(alphas, alphas + nums + "_$")
        mod = Literal("%")
        plus = Literal("+")
        minus = Literal("-")
        mult = Literal("*")
        iadd = Literal("+=")
        imult = Literal("*=")
        idiv = Literal("/=")
        isub = Literal("-=")
        div = Literal("/")
        lpar = Literal("(").suppress()
        rpar = Literal(")").suppress()
        addop = plus | minus
        multop = mult | div | mod
        iop = iadd | isub | imult | idiv
        expop = Literal("^")
        pi = CaselessLiteral("PI")
        expr = Forward()
        atom = (
            (
                Optional(oneOf("- +"))
                + (ident + lpar + expr + rpar | pi | e | fnumber).setParseAction(self.pushFirst)
            )
            | Optional(oneOf("- +")) + Group(lpar + expr + rpar)
        ).setParseAction(self.pushUMinus)
        # by defining exponentiation as "atom [ ^ factor ]..." instead of
        # "atom [ ^ atom ]...", we get right-to-left exponents, instead of left-to-right
        # that is, 2^3^2 = 2^(3^2), not (2^3)^2.
        factor = Forward()
        factor << atom + ZeroOrMore((expop + factor).setParseAction(self.pushFirst))
        term = factor + ZeroOrMore((multop + factor).setParseAction(self.pushFirst))
        expr << term + ZeroOrMore((addop + term).setParseAction(self.pushFirst))
        final = expr + ZeroOrMore((iop + expr).setParseAction(self.pushFirst))
        # addop_term = ( addop + term ).setParseAction( self.pushFirst )
        # general_term = term + ZeroOrMore( addop_term ) | OneOrMore( addop_term)
        # expr <<  general_term
        self.bnf = final
        # map operator symbols to corresponding arithmetic operations
        epsilon = 1e-12
        self.opn = {
            "+": operator.add,
            "-": operator.sub,
            "+=": operator.iadd,
            "-=": operator.isub,
            "*": operator.mul,
            "*=": operator.imul,
            "/": operator.truediv,
            "/=": operator.itruediv,
            "^": operator.pow,
            "%": operator.mod,
        }
        self.fn = {
            "sin": math.sin,
            "cos": math.cos,
            "tan": math.tan,
            "exp": math.exp,
            "abs": abs,
            "trunc": lambda a: int(a),
            "round": round,
            "sgn": lambda a: abs(a) > epsilon and ((a > 0) - (a < 0)) or 0,
            "log": lambda a: math.log(a, 10),
            "ln": math.log,
            "log2": math.log2,
            "sqrt": math.sqrt,
        }

    def evaluateStack(self, s):
        op = s.pop()
        if op == "unary -":
            return -self.evaluateStack(s)
        if op in self.opn:
            op2 = self.evaluateStack(s)
            op1 = self.evaluateStack(s)
            return self.opn[op](op1, op2)
        elif op == "PI":
            return math.pi  # 3.1415926535
        elif op == "E":
            return math.e  # 2.718281828
        elif op in self.fn:
            return self.fn[op](self.evaluateStack(s))
        elif op[0].isalpha():
            return 0
        else:
            return float(op)

    def eval(self, num_string, parseAll=True):
        self.exprStack = []
        results = self.bnf.parseString(num_string, parseAll)
        return self.evaluateStack(self.exprStack[:])


NSP = NumericStringParser()


class MathBlock(Block):
    ACCEPTED_NAMES = ("math", "m", "+", "calc")

    def process(self, ctx: Context):
        try:
            return str(NSP.eval(ctx.verb.payload.strip(" ")))
        except:
            return None
